#ifdef _WIN32
#include <windows.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include <mmsystem.h>
#include "gl\glut.h"
#else
#include <sys/time.h>
#include <stdlib.h>
#ifdef __UBUNTU__
#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glut.h>
#else
#include <opengl/gl.h>
#include <opengl/glu.h>
#include <GLUT/GLUT.h>
#endif
#define TRUE 1
#define FALSE 0
#define BOOL int
#endif

#include <math.h>
#include "timer.h"
#include "cubegrid.h"

BOOL DemoInit();
BOOL GLInit(int, int);
void UpdateFrame();

// Marching Cubes lattice dimensions
const int GridSize=60;
// the ISO-surface value
float threshold=1.0f;

typedef struct {  // structure to represent a metaball
 vector p;
 float r2;
} MetaBall;

#define numMetaballs  4

// The list of MataBalls
MetaBall metaballs[numMetaballs];
// The cubegrid structire 
CubeGrid cubeGrid={0,NULL,0,NULL};


void MetaBallInit(MetaBall *m, float x, float y, float z, float r2){
  m->p.x=x;
  m->p.y=y;
  m->p.z=z;
  m->r2=r2;
}

BOOL DemoInit(){
  int i;
 if(!CubeGrid_CreateMemory(&cubeGrid))return FALSE;
 if(!CubeGrid_Init(&cubeGrid,GridSize))return FALSE;
 for(i=0; i<numMetaballs; i++)MetaBallInit(&metaballs[i],0.0f, 0.0f, 0.0f, 5.0f+(float)i);
 Time_Reset();
 return TRUE;
}

void UpdateFrame(){
 //update the metaballs position and calculate surface normal vectors
 vector b2p; // /from ball centre
 float r2;
 vector p;  // working copy of metaball centre
 float scale,d2;
  int i,j;
 float c = 2.0f*(float)cos(Time_GetTime()/600);
 // position the metaballs - they all lie in one plane
 metaballs[0].p.x=-4.0f*(float)cos(Time_GetTime()/700) - c;
 metaballs[0].p.y=4.0f*(float)sin(Time_GetTime()/600) - c;
 metaballs[1].p.x=5.0f*(float)sin(Time_GetTime()/400) + c;
 metaballs[1].p.y=5.0f*(float)cos(Time_GetTime()/400) - c;
 metaballs[2].p.x=-5.0f*(float)cos(Time_GetTime()/400) - 0.2f*(float)sin(Time_GetTime()/600);
 metaballs[2].p.y=5.0f*(float)sin(Time_GetTime()/500) - 0.2f*(float)sin(Time_GetTime()/400);

 metaballs[3].p.x=6.0f*(float)sin(Time_GetTime()/900) + c;
 metaballs[3].p.y=6.0f*(float)cos(Time_GetTime()/900) - c;
 //clear the field
 for(i=0; i<cubeGrid.nVertices; i++) {
  cubeGrid.vertices[i].v=0.0f;
  cubeGrid.vertices[i].n.x=0.0;
  cubeGrid.vertices[i].n.y=0.0;
  cubeGrid.vertices[i].n.z=0.0;
 }
 //evaluate the scalar field at each point
 for(i=0; i<numMetaballs; i++){
  r2=metaballs[i].r2;
  p=metaballs[i].p;
  for(j=0; j<cubeGrid.nVertices; j++) {
   b2p.x=cubeGrid.vertices[j].p.x-p.x;
   b2p.y=cubeGrid.vertices[j].p.y-p.y;
   b2p.z=cubeGrid.vertices[j].p.z-p.z;
   //get squared distance from metaball ball centre to cube cordner
   d2 = b2p.x*b2p.x + b2p.y*b2p.y + b2p.z*b2p.z;
   if(d2==0.0f)d2=0.0001f;
   cubeGrid.vertices[j].v+=r2/d2;
   scale=r2/(d2*d2);
   cubeGrid.vertices[j].n.x+=b2p.x*scale;
   cubeGrid.vertices[j].n.y+=b2p.y*scale;
   cubeGrid.vertices[j].n.z+=b2p.z*scale;
  }
 }
}


void drawTriangleVertex(GLfloat *n, GLfloat *v){
 // called by the marching cube algorithm 3 times to draw each triangle vertex
 glNormal3fv(n);
 glVertex3fv(v);
} 

void display(void) {
 UpdateFrame();
 glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
 glLoadIdentity();  
 glEnable(GL_LIGHTING);
 glTranslatef(0.0f, 0.0f, -30.0f);
 glRotatef((float)Time_GetTime()/30, 1.0f, 0.0f, 1.0f);
 glBegin(GL_TRIANGLES); 
 CubeGrid_DrawSurface(&cubeGrid,threshold); // use marching cubes to draw the surface
 glEnd();
 glDisable(GL_LIGHTING);
 glutSwapBuffers();
}

BOOL GLInit(int ww, int hh){
 int height;
 GLfloat ambient[]={0.1f, 0.5f, 0.1f, 1.0f};
 GLfloat position[]={-1.0f, 1.0f, 1.0f, 0.0f};
 GLfloat specular[]={1.0f, 1.0f, 1.0f, 1.0f};
 GLfloat diffuse[]={0.2f, 0.8f, 0.2f, 1.0f};
 GLfloat white[]={1.0f, 1.0f, 1.0f, 1.0f};
 float shininess=32.0f;
 height=hh;
 glViewport(0, 0, ww, height);
 glMatrixMode(GL_PROJECTION);
 glLoadIdentity();
 gluPerspective(45.0f, (GLfloat)ww/(GLfloat)height, 1.0f, 100.0f);
 glMatrixMode(GL_MODELVIEW);
 glLoadIdentity();
 glShadeModel(GL_SMOOTH);
 glClearColor( 0.1,0.1,0.1,1.0);
 glColor4f(1.0f, 1.0f, 1.0f, 1.0f);
 glClearDepth(1.0f);
 glEnable(GL_DEPTH_TEST);
 glDepthFunc(GL_LEQUAL);
 glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
 glEnable(GL_NORMALIZE);
 glEnable(GL_CULL_FACE);
 glLightfv(GL_LIGHT1, GL_AMBIENT, ambient);
 glLightfv(GL_LIGHT1, GL_DIFFUSE, diffuse);
 glLightfv(GL_LIGHT1, GL_POSITION, position);
 glLightfv(GL_LIGHT1, GL_SPECULAR, specular);
 glEnable(GL_LIGHT1);
 glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, white);
 glMaterialfv(GL_FRONT, GL_SPECULAR, white);
 glMaterialfv(GL_FRONT, GL_SHININESS, &shininess);
 return TRUE;
}

void Keyboard(unsigned char key, int x, int y){
 if(key == 'w')glPolygonMode(GL_FRONT, GL_LINE);  // wireframe 
 if(key == 's')glPolygonMode(GL_FRONT, GL_FILL);  // solid shading
}

void reshape(GLint width, GLint height){
   glViewport(0, 0, width, height);
   glMatrixMode(GL_PROJECTION);
   glLoadIdentity();
   gluPerspective(45.0, (float)width / height, 1.0f, 100.0f);
   glMatrixMode(GL_MODELVIEW);
}

int main(int argc, char* argv[]) {
  int w=1024,h=1024;
  int window_id;
  glutInit(&argc, argv);
  glutInitWindowSize(w,h);
  glutInitDisplayMode( GLUT_RGBA | GLUT_DOUBLE | GLUT_DEPTH );
  window_id = glutCreateWindow( "Marching Cubes / Metaballs" );
  glutSetWindow( window_id );
  glutDisplayFunc(display);
  glutReshapeFunc(reshape);
  glutKeyboardFunc(Keyboard);
  glutIdleFunc(display);
  DemoInit();
  GLInit(w,h);
  glutMainLoop();
  return 0;
}
